#ifndef CLASS_H_INCLUDED_
#define CLASS_H_INCLUDED_

#include <string>
#include <vector>

#include "rose.h"

#include "object.h"
#include "definition.h"

namespace risc {

class Event;
class EventAndList;
class EventOrList;
class Function;
class HierarchicalChannelInstance;
class InOutPort;
class InPort;
class Instance;
class Module;
class ModuleInstance;
class OutPort;
class Port;
class PrimitiveChannelInstance;
class Variable;

/*TLM 2.0*/
class Socket;
class InitiatorSocket;
class TargetSocket;

typedef std::vector<Event*> EventVector;
typedef std::vector<EventAndList*> EventAndListVector;
typedef std::vector<EventOrList*> EventOrListVector;
typedef std::vector<Function*> FunctionVector;
typedef std::vector<HierarchicalChannelInstance*>
            HierarchicalChannelInstanceVector;
typedef std::vector<InOutPort*> InOutPortVector;
typedef std::vector<InPort*> InPortVector;
typedef std::vector<ModuleInstance*> ModuleInstanceVector;
typedef std::vector<OutPort*> OutPortVector;
typedef std::vector<Port*> PortVector;
typedef std::vector<PrimitiveChannelInstance*> PrimitiveChannelInstanceVector;
typedef std::vector<Variable*> VariableVector;

/*TLM 2.0*/
typedef std::vector<Socket*> SocketVector;
typedef std::vector<InitiatorSocket*>InitiatorSocketVector;
typedef std::vector<TargetSocket*>TargetSocketVector;

class Class: public Definition {

public:

  /**
   * \brief Default constructor
   */
  explicit Class(SgClassDefinition *ast_node, Type type);

  ~Class();

  virtual void initialize_functions();

  /**
   * \brief  This function searches for all variable declarations in the module
   *         definition. The corresponding variable vectors will be used.
   */
  void initialize_variables();

  /**
   * \fn       get_before_end_of_elaboration_definition
   * \brief    This function return a pointer to the definition of the function
   *           before_end_of_elaboration
   */
  SgFunctionDefinition* get_before_end_of_elaboration_definition();

  /**
   * \brief Find a port by it's ast variable definition
   */
  Variable* find_variable(SgVariableDefinition* var_def);

  /**
   * \brief Find a port by it's ast variable definition
   */
  Port* find_port(SgVariableDefinition* var_def);

  /**
   * \brief  Find an event by it's ast variable definition
   */
  Event* find_event(SgVariableDefinition* var_def);

  /**
   * \brief  Find an event by it's ast variable definition
   */
  EventAndList* find_event_and_list(SgVariableDefinition* var_def);

  /**
   * \brief  Find an event by it's ast variable definition
   */
  EventOrList* find_event_or_list(SgVariableDefinition* var_def);

  /**
   * \brief  Find an module by it's ast variable definition
   */
  ModuleInstance* find_module(SgVariableDefinition* var_def);

  /**
   * \brief find a prim channel by it's ast variable definition
   */
  PrimitiveChannelInstance* find_prim_channel(SgVariableDefinition* var_def);

  /**
   * \brief find a hier channel by it's ast variable definition
   */
  HierarchicalChannelInstance* find_hier_channel(SgVariableDefinition* var_def);

  /**
   * \brief find a hier channel/module/ prim channel by it's ast variable
   *        definition
   */
  Instance* find_instance(SgVariableDefinition* var_def);

  Socket* find_socket(SgVariableDefinition* var_def);
  
  /**
   * \brief find a target socket by it's ast variable
   *        definition
   */
  TargetSocket* find_target_socket(SgVariableDefinition* var_def);

  /**
   * \brief find an initiator socket by it's ast variable
   *        definition
   */
  InitiatorSocket* find_initiator_socket(SgVariableDefinition* var_def);

  /**
   * \brief  Returns the ast definition of the class
   * \return Pointer to the definition
   */
  SgClassDefinition* get_ast_node();

  /**
   * \brief  returns the name of the module definition
   */
  std::string get_name();

  /**
   * \brief  Returns the name of the module type
   * \return name of the module instance
   */
  virtual std::string get_ast_type_name();

  /**
   * \brief Vector of events which are defined in this module
   */
  EventVector events_;

  /**
   * \brief Vector of event and lists which are defined in this module
   */
  EventAndListVector event_and_lists_;

  /**
   * \brief Vector of event or lists which are defined in this module
   */
  EventOrListVector event_or_lists_;

  /**
   * \brief Vector of C++ builtin variables that are declared in the module
   */
  VariableVector variables_;

  /**
   * \brief Vector of inports which are defined in this module
   */
  InPortVector inports_;

  /**
   * \brief Vector of inoutports which are defined in this module
   */
  InOutPortVector inoutports_;

  /**
   * \brief Vector of outports which are defined in this module
   */
  OutPortVector outports_;

  /**
   * \brief Vector of initiator sockets which are defined in this module
   */
  InitiatorSocketVector initiatorsockets_;

  /**
   * \brief Vector of targets sockets which are defined in this module
   */
  TargetSocketVector targetsockets_;

  /**
   * \brief Vector of primitive channel instances which are declared in this
   *        module
   */
  PrimitiveChannelInstanceVector sub_prim_channel_;

  /**
   * \brief Vector of hierarchical channel instances which are declared in this
   *        module
   */
  HierarchicalChannelInstanceVector sub_hier_channel_;

  /**
   * \brief Vector of module instances which are declared in this module.
   */
  ModuleInstanceVector sub_modules_;

  /**
   * \brief Vector of functions which are defined in this module
   */
  FunctionVector functions_;

  /**
   * \fn     initialize_mapping
   * \brief  This function analyzes the mapping between a module/channel
   *         definition and its children
   */
  void initialize_mapping();

  /**
   * \fn    analyze_port_mapping(
      Module *parent_module, ModuleInstance *sub_module_instance, Port *port)
   * \brief This is a helper function for internal use only
   */
  void analyze_port_mapping(
      Module *parent_module, ModuleInstance *sub_module_instance, Port *port);

  /**
   * \fn    analyze_port_mapping_style3(
      Module *parent_module, ModuleInstance *sub_module_instance, Port *port)
   * \brief This is a helper function for internal use only
   */
  void analyze_port_mapping_style3(
      Module *parent_module, ModuleInstance *sub_module_instance, Port *port);

  /**
   * \fn    analyze_socket_mapping(
      Module *parent_module, ModuleInstance *sub_module_instance, Socket *socket)
   * \brief This is a helper function for internal use only
   */
  void analyze_socket_mapping(
      Module *parent_module, ModuleInstance *sub_module_instance, Socket *socket);

  /**
   * \fn    analyze_general_mapping
   * \brief This is a helper function for internal use only
   */
  void analyze_general_mapping();
  /**
   * \fn    has_module_pointers
   * \brief This functions returns true if the class has module member
   *        which is declared as a pointer.
   */
  bool has_module_pointers();

  /**
   * \fn    has_channel_pointers
   * \brief This functions returns true if the class has channel member
   *        which is declared as a pointer.
   */
  bool has_channel_pointers();

  typedef std::set<
    std::pair<
      std::list<SgVariableSymbol*>,
      std::list<SgVariableSymbol*>
    >
  > PathMappingType;
  
  PathMappingType path_mappings;

  bool is_used_;

private:
  /**
   * \brief Disabled copy constructor
   */
  Class(const Class &c);
};


} // end of namespace risc


#endif /* CLASS_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
