#ifndef CLASS_IS_DERIVED_FROM_H_INCLUDED_
#define CLASS_IS_DERIVED_FROM_H_INCLUDED_

#include <unordered_map>

#include "rose.h"


namespace risc {

namespace tools {

/**
 * \brief lookup for derived classes (base class, derived class, is derived)
 */
extern std::unordered_map<SgClassDefinition*,
                          std::unordered_map<SgClassDefinition*, bool> > lookup;

/**
 * \fn                  is_derived_from(
						    SgClassDefinition* derived_class,
						    SgClassDefinition* base_class)
 * \brief               This function checks if the derived class is derived
 *                      from the base class
 *                      Caching techniques are applied for speedup
 * \param derived_class the derived class definition
 * \param base_class    the base class definition
 * \return              true, if the class definition derived_class is derived
 *                      from calls definition base_class. Otherwise false
 */
bool
is_derived_from(
    SgClassDefinition* derived_class,
    SgClassDefinition* base_class);

/**
 * \fn                  is_derived_from_internal(
					    SgClassDefinition* derived_class,
					    SgClassDefinition* base_class)
 * \brief               This function is for internal usage only because
 *                      no caching is applied
 * \param derived_class the derived class definition
 * \param base_class    the base class definition
 * \return              true, if the class definition derived_class is derived
 *                      from calls definition base_class. Otherwise false
 */
bool
is_derived_from_internal(
    SgClassDefinition* derived_class,
    SgClassDefinition* base_class);

/**
 * \fn 					is_derived_from_sc_interface(SgClassDefinition* derived_class)
 * \brief               This function checks if the class is derived from
 *                      sc_interface
 * \param derived_class the derived class definition
 * \return              True, if the class is derived from sc_interface
 */
bool
is_derived_from_sc_interface(SgClassDefinition* derived_class);

/**
 * \brief               This function checks if the class is derived from
 *                      sc_channel
 * \param derived_class the derived class definition
 * \return              True, if the class is derived from sc_module
 */
bool
is_derived_from_sc_channel(SgClassDefinition* derived_class);

/**
 * \brief               This function checks if the class is derived from
 *                      sc_prim_channel
 * \param derived_class the derived class definition
 * \return              True, if the class is derived from sc_module
 */
bool
is_derived_from_sc_prim_channel(SgClassDefinition* derived_class);

/**
 * \brief               This function checks if the class is derived from
 *                      sc_module
 * \param derived_class the derived class definition
 * \return              True, if the class is derived from sc_module
 */
bool
is_derived_from_sc_module(SgClassDefinition* derived_class);

/**
 * \brief               This function checks if the class is derived from
 *                      sc_port
 * \param derived_class the derived class definition
 * \return              True, if the class is derived from sc_port
 */
bool
is_derived_from_sc_port(SgClassDefinition* derived_class);

/**
 * \brief               This function checks if the class is derived from
 *                      sc_port_base
 * \param derived_class the derived class definition
 * \return              True, if the class is derived from sc_port_base
 */
bool
is_derived_from_sc_port_base(SgClassDefinition* derived_class);

/**
 * \brief               This function checks if the class is derived from
 *                      tlm_base_initiator_socket_b
 * \param derived_class the derived class definition
 * \return              True, if the class is derived from tlm_base_initiator_socket_b
 */
bool
is_derived_from_tlm_base_initiator_socket_b(SgClassDefinition* derived_class);

/**
 * \brief               This function checks if the class is derived from
 *                      tlm_base_target_socket_b
 * \param derived_class the derived class definition
 * \return              True, if the class is derived from tlm_base_target_socket_b
 */
bool
is_derived_from_tlm_base_target_socket_b(SgClassDefinition* derived_class);

/**
 * \brief               This function checks if the class is derived from
 *                      sc_port_base
 * \param var_ref the   derived class definition
 * \return              True, if the class is derived from sc_port_base
 */
bool
is_derived_from_sc_port_base(SgVarRefExp* var_ref);


/**
 * \brief               This function checks if the class is derived from
 *                      sc_port_base
 * \param type the      derived class type
 * \return              True, if the class is derived from sc_port_base
 */
bool
is_derived_from_sc_port_base(SgType* type);

/**
 * \brief               This function checks if the class is derived from
 *                      sc_prim_channel
 * \param derived_class the derived class definition
 * \return              True, if the class is derived from sc_prim_channel
 */
bool
is_derived_from_sc_prim_channel(SgClassDefinition* derived_class);

/**
 * \brief               This function checks if the class is derived from
 *                      sc_channel
 * \param derived_class the derived class definition
 * \return              True, if the class is derived from sc_channel
 */
bool
is_derived_from_sc_channel(SgClassDefinition* derived_class);

/**
 * \brief               This function checks if the class is derived from
 *                      sc_event or if it is sc_event itself
 * \param derived_class the derived class definition
 * \return              True, if the class is derived from sc_event
 */
bool
is_sc_event(SgClassDefinition* derived_class);

/**
 * \brief               This function checks if the class is derived from
 *                      sc_event_and_list or if it is sc_event_and_list itself
 * \param derived_class the derived class definition
 * \return              True, if the class is derived from sc_event
 */
bool
is_sc_event_and_list(SgClassDefinition* derived_class);

/**
 * \brief               This function checks if the class is derived from
 *                      sc_event_or_list or if it is sc_event_or_list itself
 * \param derived_class the derived class definition
 * \return              True, if the class is derived from sc_event
 */
bool
is_sc_event_or_list(SgClassDefinition* derived_class);

/**
 * \brief               This function checks if the variable access is of the type
 *                      sc_time (or a reference to a sc_time)
 *
 * \param    var_ref    the variable access to check
 * \return              True, if it is the type sc_time
 */
bool
is_sc_time(SgVarRefExp* var_ref);

/**
 * \brief               This function checks if the variable access is of the type
 *                      sc_event (or a reference to a sc_event)
 *
 * \param    var_ref    the variable access to check
 * \return              True, if it is the type sc_event
 */
bool
is_sc_event(SgVarRefExp* var_ref);

/**
 * \brief               This function checks if the variable access is of the type
 *                      sc_event_and_list (or a reference to a sc_event_and_list)
 *
 * \param    var_ref    the variable access to check
 * \return              True, if it is the type sc_event_and_list
 */
bool
is_sc_event_and_list(SgVarRefExp* var_ref);

/**
 * \brief               This function checks if the variable access is of the type
 *                      sc_event_or_list (or a reference to a sc_event_and_list)
 *
 * \param    var_ref    the variable access to check
 * \return              True, if it is the type sc_event_or_list
 */
bool
is_sc_event_or_list(SgVarRefExp* var_ref);

/**
 * \brief               This function checks if the class is derived from
 *                      sc_port_base and not from tlm_base_initiator_socket_b
 * \param derived_class the derived class definition
 * \return              True, if the class is derived from sc_port_base
 */
bool
is_sc_port(SgClassDefinition* derived_class);

/**
 * \brief               This function checks if the template class bases on
 *                      the class template sc_in_port
 *
 * \param    template_class      the class definition
 * \return              True, if the class is a SystemC port
 */
bool
is_sc_in_port(SgTemplateInstantiationDefn* template_class);

/**
 * \brief               This function checks if the template class bases on
 *                      the class template sc_out_port
 *
 * \param    template_class      the class definition
 * \return              True, if the template class bases on the class
 *                      template sc_out_port
 */
bool
is_sc_out_port(SgTemplateInstantiationDefn* template_class);

/**
 * \brief               This function checks if the template class bases on
 *                      the class template sc_inout_port
 *
 * \param    template_class      the class definition
 * \return              True, if the template class bases on the class
 *                      template sc_inout_port
 */
bool
is_sc_inout_port(SgTemplateInstantiationDefn* template_class);

/**
 * \brief               This function checks if the given defintion is qual
 *                      to the definition of a sc_in_clk
 *
 * \param    definition      the class definition
 * \return              True, if equal
 */
bool
is_sc_in_clk(SgClassDefinition* definition);

/**
 * \brief               This function checks if the given defintion is qual
 *                      to the definition of a sc_out_clk
 *
 * \param    definition      the class definition
 * \return              True, if equal
 */
bool
is_sc_out_clk(SgClassDefinition* definition);

/**
 * \brief               This function checks if the given defintion is qual
 *                      to the definition of a sc_inout_clk
 *
 * \param    definition      the class definition
 * \return              True, if equal
 */
bool
is_sc_inout_clk(SgClassDefinition* definition);

/**
 * \brief               This function checks if the given defintion is qual
 *                      to the definition of a sc_in<bool>
 *
 * \param    definition      the class definition
 * \return              True, if equal
 */
bool
is_sc_in_bool(SgClassDefinition* definition);

/**
 * \brief               This function checks if the variable is of the type
 *                      sc_fifo_in
 *
 * \param    var_ref    the variable to check
 * \return              True, if it is the type sc_fifo_in
 */
bool
is_sc_fifo_in(SgVarRefExp *var_ref);

/**
 * \brief               This function checks if the variable is of the type
 *                      sc_fifo_out
 *
 * \param    var_ref    the variable to check
 * \return              True, if it is the type sc_fifo_out
 */
bool
is_sc_fifo_out(SgVarRefExp *var_ref);

} // end namespace tools

} // end namespace risc

#endif /* CLASS_IS_DERIVED_FROM_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
