// delta_pipe9.cc: delta-cycle pipeline with 9 inner stages
// 02/13/20, RD

#include <stdio.h>
#include <assert.h>

#include "systemc.h"

#define STAGES     9
#define ITEMS      10
#define WORKLOAD   10000000*STAGES
//#define VERBOSE

#pragma RISC M conflict-free
int M[STAGES+1][2] = { }; // memory for double-buffering

unsigned long Work(int Load)
{
    int Count;
    unsigned long Result = 42;

    for(Count=0; Count<Load; Count++)
    {
	Result *= 17;
    }
    return Result;
}

SC_MODULE(Source)
{
    int Buffer, Index, Count;
    sc_event &OutDataReady, &OutBufferClear;

    void main(void)
    {
	// send data item 1
	M[Buffer][Index] = ++Count;
#ifdef VERBOSE
	printf("%2u: %s sends data %d\n", (unsigned)sc_delta_count(), name(), Count);
#endif
	OutDataReady.notify(SC_ZERO_TIME);
	wait(SC_ZERO_TIME);

	// send data item 2, 3, 4, ...
	for(int i=1; i<ITEMS; i++)
	{
	    M[Buffer][Index^=1] = ++Count;
#ifdef VERBOSE
	    printf("%2u: %s sends data %d\n", (unsigned)sc_delta_count(), name(), Count);
#endif
	    OutDataReady.notify(SC_ZERO_TIME);
	    wait(OutBufferClear);
	}
	assert(Count == ITEMS);
	assert(sc_time_stamp() == SC_ZERO_TIME);
	assert(sc_delta_count() == ITEMS);
    }

    Source(sc_module_name _nm, int _b,
		sc_event &_dr, sc_event &_bc)
    :	sc_module(_nm), Buffer(_b),
	OutDataReady(_dr), OutBufferClear(_bc)
    {
	Index = 0;
	Count = 0;
	SC_HAS_PROCESS(Source);
	SC_THREAD(main);
    }
};

SC_MODULE(Stage)
{
    int InBuffer, OutBuffer, Index, d;
    sc_event &InDataReady, &InBufferClear;
    sc_event &OutDataReady, &OutBufferClear;
    unsigned long w;

    void main(void)
    {
	// wait for data
	wait(InDataReady);

	// process data item 1
	d = M[InBuffer][Index];
	w = Work(WORKLOAD);
	M[OutBuffer][Index] = d;
#ifdef VERBOSE
	printf("%2u: %s processes data %d\n", (unsigned)sc_delta_count(), name(), d);
#endif
	InBufferClear.notify(SC_ZERO_TIME);
	OutDataReady.notify(SC_ZERO_TIME);
	wait(InDataReady);

	// process data item 2, 3, 4, ...
	for(;;)
	{
	    d = M[InBuffer][Index^=1];
	    w = Work(WORKLOAD);
	    M[OutBuffer][Index] = d;
#ifdef VERBOSE
	    printf("%2u: %s processes data %d\n", (unsigned)sc_delta_count(), name(), d);
#endif
	    InBufferClear.notify(SC_ZERO_TIME);
	    OutDataReady.notify(SC_ZERO_TIME);
	    wait(InDataReady & OutBufferClear);
	}
    }

    Stage(sc_module_name _nm, int _ib, int _ob,
		sc_event &_idr, sc_event &_ibc,
		sc_event &_odr, sc_event &_obc)
    :	sc_module(_nm), InBuffer(_ib), OutBuffer(_ob),
	InDataReady(_idr), InBufferClear(_ibc),
	OutDataReady(_odr), OutBufferClear(_obc)
    {
	Index = 0;
	SC_HAS_PROCESS(Stage);
	SC_THREAD(main);
    }
};

SC_MODULE(Sink)
{
    int Buffer, Index, d, Count;
    sc_event &InDataReady, &InBufferClear;

    void main(void)
    {
	// receive data items 1, 2, 3, ...
	for(int i=0; i<ITEMS; i++)
	{
	    wait(InDataReady);
	    d = M[Buffer][Index];
#ifdef VERBOSE
	    printf("%2u: %s received data %d\n", (unsigned)sc_delta_count(), name(), d);
#endif
	    InBufferClear.notify(SC_ZERO_TIME);
	    Index^=1;
	    Count++;
	    assert(d == Count);
	}
	assert(Count == ITEMS);
	assert(sc_time_stamp() == SC_ZERO_TIME);
	assert(sc_delta_count() == ITEMS+STAGES);
    }

    Sink(sc_module_name _nm, int _b,
		sc_event &_dr, sc_event &_bc)
    :	sc_module(_nm), Buffer(_b),
	InDataReady(_dr), InBufferClear(_bc)
    {
	Index = 0;
	Count = 0;
	SC_HAS_PROCESS(Sink);
	SC_THREAD(main);
    }
};

SC_MODULE(TOP)
{
    sc_event data_ready1, buffer_clear1;
    sc_event data_ready2, buffer_clear2;
    sc_event data_ready3, buffer_clear3;
    sc_event data_ready4, buffer_clear4;
    sc_event data_ready5, buffer_clear5;
    sc_event data_ready6, buffer_clear6;
    sc_event data_ready7, buffer_clear7;
    sc_event data_ready8, buffer_clear8;
    sc_event data_ready9, buffer_clear9;
    sc_event data_ready10, buffer_clear10;

    Source src;
    Stage  s1, s2, s3, s4, s5, s6, s7, s8, s9;
    Sink   snk;

    SC_CTOR(TOP)
    : src("src", 0, data_ready1, buffer_clear1),
	s1("s1", 0, 1, data_ready1, buffer_clear1, data_ready2, buffer_clear2),
	s2("s2", 1, 2, data_ready2, buffer_clear2, data_ready3, buffer_clear3),
	s3("s3", 2, 3, data_ready3, buffer_clear3, data_ready4, buffer_clear4),
	s4("s4", 3, 4, data_ready4, buffer_clear4, data_ready5, buffer_clear5),
	s5("s5", 4, 5, data_ready5, buffer_clear5, data_ready6, buffer_clear6),
	s6("s6", 5, 6, data_ready6, buffer_clear6, data_ready7, buffer_clear7),
	s7("s7", 6, 7, data_ready7, buffer_clear7, data_ready8, buffer_clear8),
	s8("s8", 7, 8, data_ready8, buffer_clear8, data_ready9, buffer_clear9),
	s9("s9", 8, 9, data_ready9, buffer_clear9, data_ready10, buffer_clear10),
	snk("snk", 9, data_ready10, buffer_clear10)
    {
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    printf("delta_pipe9 starts\n");
    sc_start();
    assert(top.src.Count == ITEMS);
    assert(top.snk.Count == ITEMS);
    assert(sc_time_stamp() == SC_ZERO_TIME);
    printf("delta_pipe9 done.\n");
    printf("[potential speedup over SEQ: OOO 9x, NPD 9x, SYN 9x]\n");
    return 0;
}

// EOF
